//////////////////////////////////////////////////////////
// Main CPU routines

#ifdef MSS
#include <mss.h>
#endif

#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <ctype.h>

#include "vb_types.h"
#include "v810_ins.h"
#include "v810_opt.h"
#include "v810_cpu.h"
#include "v810cpuD.h"
#include "vb_vbt.h"
#include "vb_vbtD.h"
//Options...
#include "vb_set.h"

////////////////////////////////////////////////////////////
// Globals
WORD P_REG[32];  // Main program reg pr0-pr31
WORD S_REG[32];  // System registers sr0-sr31
WORD PC;         // Program Counter

// Reinitialize the defaults in the CPU
void v810_reset() {
    P_REG[0]      =  0x00000000;
    PC            =  0xFFFFFFF0;
    S_REG[ECR]    =  0x0000FFF0;
    S_REG[PSW]    =  0x00008000;
    S_REG[PIR]    =  0x00008300;
    S_REG[TKCW]   =  0x000000E0;
    //S_REG[HCCW] =  0x0000FFF0; //V830?
}

//could be done a lot better, later maby!
void v810_trc(int num, int noKey) {
    int      lowB, highB, lowB2, highB2;             // up to 4 bytes for instruction (either 16 or 32 bits)
    int      opcode;
	int arg1 = 0;
	int arg2 = 0;
	int arg3 = 0;
    int Ttmp, tmp2;
    //WORD tmpWord;
    HWORD tmpHWordL;
    HWORD tmpHWordH;
    int i, j;
    int step;
    int flags = 0;
    INT64 temp = 0;
    INT64U tempu = 0;
    int val = 0;
    WORD msb = 0;
    int tCount = 0;

    for(i = 0; i< num; i++) {
/*        if(!noKey) {
            if(kbhit())
                if(getch() == 27) break; //Fix with Allegro!
        }     */

        //Adjust the Timer.... should not be hear!
        //Many bugs, look into it more...
        if(tHReg.TCR & 0x01) { // Timer Enabled
		step = 1;
		if(tHReg.TCR & 0x10) step = 5;

            Ttmp = tHReg.TLB + (tHReg.THB << 8);
            Ttmp-=step;
            tHReg.TLB = (Ttmp&0xFF);
            tHReg.THB = ((Ttmp >> 8)&0xFF);

            if(Ttmp < step) {
                tHReg.TCR = tHReg.TCR | 0x02; //Zero Status
                if(tHReg.TCR & 0x08) { //Timer Interupt Enabled
                    //if(tVIPREG.INTENB&0xFFFF) { //Wrong Number???
                        v810_int(1); 
                    //}
                }
            }
        }

        //PC     = (((PC&0x07FFFFFE) & V810_ROM1.highaddr)|0x07000000); //If PC goes belowe 0x07000000 than we have a problem!!!
        tmp2 = PC;
        PC     = ((PC&0x07FFFFFE) & V810_ROM1.highaddr);

        // Interactive Dissasemble, remove in release...
        if(tVBOpt.DISASM) { //turn on and off
            v810_addDasm(PC);
        }

        //Special Stack Trace (of sourts), remove in release
        vb_addQueue(PC); // Circular queue of last 100 instructions
        if(PC < 0x07000000) {
            dtprintf(6,ferr,"\nPC Invalid %4x",tmp2); //If PC goes belowe 0x07000000 than we have a problem!!!
            vb_dumpQueue();
            fclose(ferr);
            exit(1);
        }
        lowB   = ((BYTE *)(V810_ROM1.off + PC))[0];
        highB  = ((BYTE *)(V810_ROM1.off + PC))[1];
        lowB2  = ((BYTE *)(V810_ROM1.off + PC))[2];
        highB2 = ((BYTE *)(V810_ROM1.off + PC))[3];


        //Remove Me!!!
        P_REG[0]=0; //Zero the Zero Reg!!!

        opcode = highB >> 2;
        if((highB & 0xE0) == 0x80)        // Special opcode format for          
            opcode = (highB >> 1);    // type III instructions.

        if((opcode > 0x4F) || (opcode < 0)) {
            //Error Invalid opcode!
            dtprintf(6,ferr,"\n%08lx\t\t%2x %2x  ;Invalid Opcode", PC, lowB, highB);
            PC += 2;                                                
            continue;                                 
        }
        
        switch(optable[opcode].addr_mode) {
          case AM_I:       // Do the same Ither way =)
            arg1 = (lowB & 0x1F);
            arg2 = (lowB >> 5) + ((highB & 0x3) << 3);
            PC += 2;   // 16 bit instruction
            break;
          case AM_II:
            arg1 = (lowB & 0x1F);
            arg2 = (lowB >> 5) + ((highB & 0x3) << 3);
            PC += 2;   // 16 bit instruction
            break;
          case AM_III:
            arg1 = ((highB & 0x1) << 8) + (lowB & 0xFE);
            break;
          case AM_IV:
            arg1 = ((highB & 0x3) << 24) + (lowB << 16) + (highB2 << 8) + lowB2;
            break;
          case AM_V:       
            arg3 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg1 = (highB2 << 8) + lowB2;
            PC += 4;   // 32 bit instruction
            break;
          case AM_VIa:  // Mode6 form1
            arg1 = (highB2 << 8) + lowB2;
            arg2 = (lowB & 0x1F);
            arg3 = (lowB >> 5) + ((highB & 0x3) << 3);
            PC += 4;   // 32 bit instruction
            break;
          case AM_VIb:  // Mode6 form2
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (highB2 << 8) + lowB2;                              //  whats the order??? 2,3,1 or 1,3,2
            arg3 = (lowB & 0x1F);
            PC += 4;   // 32 bit instruction
            break;
          case AM_VII:   // Unhandled
            dtprintf(6,ferr,"\n%08lx\t\t%2x %2x %2x %2x", PC, lowB, highB, lowB2, highB2);
            PC +=4; // 32 bit instruction
            break;
          case AM_VIII:  // Unhandled
            dtprintf(6,ferr,"\n%08lx\t%\t2x %2x %2x %2x", PC, lowB, highB, lowB2, highB2);
            PC += 4;   // 32 bit instruction
            break;
          case AM_IX:
            arg1 = (lowB & 0x1); // Mode ID, Ignore for now
            PC += 2;   // 16 bit instruction
            break;
          case AM_BSTR:  // Bit String Subopcodes
            dtprintf(6,ferr,"\nBitStr");
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            PC += 2;   // 16 bit instruction
            break;
          case AM_FPP:   // Floating Point Subcode
            //dtprintf(6,ferr,"\nFPU");
            arg1 = (lowB >> 5) + ((highB & 0x3) << 3);
            arg2 = (lowB & 0x1F);
            arg3 = ((highB2 >> 2)&0x3F);
            PC += 4;   // 32 bit instruction
            break;
          case AM_UDEF:  // Invalid opcode.
          default:           // Invalid opcode.
            dtprintf(6,ferr,"\n%08lx\t\t%2x %2x  ;Invalid Opcode", PC, lowB, highB);
            PC += 2;                                                
            break;
            continue;
        }


            switch(opcode) {
              case MOV:
                P_REG[arg2] = P_REG[arg1];
                break;
              case ADD:
                flags = 0;
                temp = P_REG[arg2] + P_REG[arg1];
                // Set Flags
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if(temp != ((long)temp)) flags = flags | PSW_CY; // Old
                if(((P_REG[arg2]^(~P_REG[arg1]))&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                P_REG[arg2] = (long)temp;
                break;
              case SUB:
                flags = 0;
                temp = P_REG[arg2] - P_REG[arg1];
                // Set Flags
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if(((P_REG[arg2]^P_REG[arg1])&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;
                if(temp != ((long)temp)) flags = flags | PSW_CY; // & 0x80000000

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                P_REG[arg2] = (long)temp;
                break;
              case CMP:
                flags = 0;
                temp = P_REG[arg2]-P_REG[arg1];
                // Set Flags
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if(((P_REG[arg2]^P_REG[arg1])&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;
                if(temp != ((long)temp)) flags = flags | PSW_CY;

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case SHL:
                flags = 0;
                val = P_REG[arg1] & 0x1F;
                // set CY before we destroy the regisrer info....
                if((val != 0)&&(P_REG[arg2] >> (32 - val))&0x01) flags = flags | PSW_CY;
                P_REG[arg2] = P_REG[arg2] << val;
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case SHR:
                flags = 0; 
                val = P_REG[arg1] & 0x1F;
                // set CY before we destroy the regisrer info....
                if((val != 0)&&(P_REG[arg2] >> (val-1))&0x01) flags = flags | PSW_CY;
                P_REG[arg2] = P_REG[arg2] >> val;
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case JMP:
                PC = (P_REG[arg1] & 0xFFFFFFFE);
                break;
              case SAR:
                flags = 0;
                val = P_REG[arg1] & 0x1F;
                msb = P_REG[arg2] & 0x80000000; // Grab the MSB

            for(j = 0; j < val; j++) {  
                    // set CY before we destroy the regisrer info....
                    if((val != 0)&&(P_REG[arg2] >> (val-1))&0x01) flags = flags | PSW_CY; //Makes no sence to me =)

                    P_REG[arg2] = (P_REG[arg2] >> 1)|msb; // Apend the MSB to the end
                }
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case MUL:
                flags=0;
                temp = (INT64)P_REG[arg1] * (INT64)P_REG[arg2];
                P_REG[30]   = temp >> 32;
                // Set Flags
                if((P_REG[30]!=0x00000000)&&(P_REG[30]!=0xFFFFFFFF)&&(arg2!=30)) flags = flags |PSW_OV;
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;

                P_REG[arg2] = (long)temp;
                break;
              case DIV:
                flags = 0;
                if(P_REG[arg1] == 0) { // Div by zero error
                    // Generate exception!
                } else {
                    if((P_REG[arg2]==0x80000000)&&(P_REG[arg1]==0xFFFFFFFF)) flags = flags |PSW_OV;
                    P_REG[30]   = (long)P_REG[arg2] % (long)P_REG[arg1];
                    P_REG[arg2] = (long)P_REG[arg2] / (long)P_REG[arg1];
                    // Set Flags
                    if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                    if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                    S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags; 
                }
                break;
              case MULU:
                flags = 0;
                tempu = (INT64U)P_REG[arg1] * (INT64U)P_REG[arg2];
                P_REG[30]   = tempu >> 32;
                // Set Flags
                if((P_REG[30]!=0x00000000)&&(arg2 != 30)) flags = flags |PSW_OV;
                if ((long)tempu == 0) flags = flags | PSW_Z;
                if ((long)tempu & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                P_REG[arg2] = (WORD)tempu;
                break;
              case DIVU:
                flags = 0;
                if(P_REG[arg1] == 0) { // Div by zero error
                    // Generate exception!
                } else {
                    P_REG[30]   = (WORD)P_REG[arg2] % (WORD)P_REG[arg1];
                    P_REG[arg2] = (WORD)P_REG[arg2] / (WORD)P_REG[arg1];
                    // Set Flags
                    if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                    if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                    S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                }
                break;
              case OR:
                flags = 0;
                P_REG[arg2] = P_REG[arg1] | P_REG[arg2];
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case AND:
                flags = 0;
                P_REG[arg2] = P_REG[arg1] & P_REG[arg2];
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case XOR:
                flags = 0;
                P_REG[arg2] = P_REG[arg1] ^ P_REG[arg2];
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case NOT:
                flags = 0;
                P_REG[arg2] = 0xFFFFFFFF - P_REG[arg1];
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case MOV_I:
                P_REG[arg2] = sign_5(arg1);
                break;
              case ADD_I:
                flags = 0;
                temp = P_REG[arg2] + sign_5(arg1);
                // Set Flags
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if(((P_REG[arg2]^(~sign_5(arg1)))&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;
                if(temp != ((long)temp)) flags = flags | PSW_CY; // & 0x80000000

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                P_REG[arg2] = (WORD)temp;
                break;
              case SETF:
            //dtprintf(6,ferr,"\nUnhandled opcode! setf");
                break;
              case CMP_I:
                flags = 0;
                temp = P_REG[arg2]-sign_5(arg1);
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if(((P_REG[arg2]^(sign_5(arg1)))&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;
                if(temp != ((long)temp)) flags = flags | PSW_CY;

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case SHL_I:
                flags = 0;
                if((arg1)&&(P_REG[arg2] >> (32 - arg1))&0x01) flags = flags | PSW_CY;
                // set CY before we destroy the regisrer info....
                P_REG[arg2] = P_REG[arg2] << arg1;
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case SHR_I:
                flags = 0;
                if((arg1)&&(P_REG[arg2] >> (arg1-1))&0x01) flags = flags | PSW_CY;
                // set CY before we destroy the regisrer info....
                P_REG[arg2] = P_REG[arg2] >> arg1;
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case EI:
                S_REG[PSW] = S_REG[PSW] & (0xFFFFFFFF - PSW_ID);
                break;
              case SAR_I:
                flags = 0;
                msb = P_REG[arg2] & 0x80000000; // Grab the MSB

                //for(i = 0; i < arg1; i++) {  //Fix for Mario Tennis, whats the proper solution!
                for(j = 0; j < arg1; j++) {  
                    // set CY before we destroy the regisrer info....
                    if((arg1)&&(P_REG[arg2] >> (arg1-1))&0x01) flags = flags | PSW_CY;
                    //Shift!
                    P_REG[arg2] = (P_REG[arg2] >> 1) | msb; //Keep sticking the msb on the end
                }
                // Set Flags
                if (!P_REG[arg2]) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                break;
              case TRAP:
            dtprintf(6,ferr,"\nUnhandled opcode! trap");
                break;
              case LDSR:
            //dtprintf(6,ferr,"\nPC %04x ldsr P$%x -> S$%x: %04x",(PC-2),(arg2 & 0x1F),(arg1 & 0x1F),P_REG[(arg2 & 0x1F)]);
            S_REG[(arg1 & 0x1F)] = P_REG[(arg2 & 0x1F)];
                break;
              case STSR:
            //dtprintf(6,ferr,"\nPC %04x stsr S$%x -> P$%x: %04x",(PC-2),(arg1 & 0x1F),(arg2 & 0x1F),P_REG[(arg1 & 0x1F)]);
            P_REG[(arg2 & 0x1F)] = S_REG[(arg1 & 0x1F)];
                break;
              case DI:
                S_REG[PSW] = S_REG[PSW] | PSW_ID;
            break;
              case BV:
                if(S_REG[PSW]&PSW_OV) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BL:
                if(S_REG[PSW]&PSW_CY) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BE:
                if(S_REG[PSW]&PSW_Z) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BNH:
                if((S_REG[PSW]&PSW_Z)||(S_REG[PSW]&PSW_CY)) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BN:
                if(S_REG[PSW]&PSW_S) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BR:
                PC += (sign_9(arg1) & 0xFFFFFFFE);
                break;
              case BLT:
                if((!!(S_REG[PSW]&PSW_S))^(!!(S_REG[PSW]&PSW_OV))) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BLE:
                if(((!!(S_REG[PSW]&PSW_S))^(!!(S_REG[PSW]&PSW_OV)))||(S_REG[PSW]&PSW_Z)) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BNV:
                if(!(S_REG[PSW]&PSW_OV)) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BNL:
                if(!(S_REG[PSW]&PSW_CY)) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BNE:
                if((S_REG[PSW]&PSW_Z) == PSW_Z) {
                    PC +=2;
                } else {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                }
                break;
              case BH:
                if(!((S_REG[PSW]&PSW_Z)||(S_REG[PSW]&PSW_CY))) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BP:
                if(!(S_REG[PSW] & PSW_S)) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case NOP:
                //Its a NOP do nothing =)
                PC +=2;
                break;
              case BGE:
                if(!((!!(S_REG[PSW]&PSW_S))^(!!(S_REG[PSW]&PSW_OV)))) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
                }
                break;
              case BGT:
                if(!(((!!(S_REG[PSW]&PSW_S))^(!!(S_REG[PSW]&PSW_OV)))||(S_REG[PSW]&PSW_Z))) {
                    PC += (sign_9(arg1) & 0xFFFFFFFE);
                } else {
                    PC +=2;
              }
            break;
              case JR:
                //if(arg1 & 0x02000000) arg1 +=0xFC000000;
                PC += (sign_26(arg1) & 0xFFFFFFFE);
                break;
              case JAL:
                P_REG[31]=PC+4;
                PC += (sign_26(arg1) & 0xFFFFFFFE);
                break;
              case MOVEA:
                P_REG[arg3] = P_REG[arg2] + sign_16(arg1);
                break;
              case ADDI:
                flags = 0;
                temp = P_REG[arg2] + sign_16(arg1);
                // Set Flags
                if ((long)temp == 0) flags = flags | PSW_Z;
                if ((long)temp & 0x80000000)  flags = flags | PSW_S;
                if (((P_REG[arg2]^(~sign_16(arg1)))&(P_REG[arg2]^temp))&0x80000000) flags = flags | PSW_OV;
                if (temp != ((long)temp)) flags = flags | PSW_CY;

                S_REG[PSW] = (S_REG[PSW] & 0xFFFFFFF0)|flags;
                P_REG[arg3] = (long)temp;
                break;
              case ORI:
                flags = 0;
                P_REG[arg3] = arg1 | P_REG[arg2];
                // Set Flags
                if (P_REG[arg3] == 0) flags = flags | PSW_Z;
                if (P_REG[arg3] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case ANDI:
                flags = 0;
                P_REG[arg3] = (arg1 & P_REG[arg2]);
                // Set Flags
                if (P_REG[arg3] == 0) flags = (flags | PSW_Z);
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case XORI:
                flags = 0;
                P_REG[arg2] = arg1 ^ P_REG[arg2];
                // Set Flags
                if (P_REG[arg2] == 0) flags = flags | PSW_Z;
                if (P_REG[arg2] & 0x80000000)  flags = flags | PSW_S;
                S_REG[PSW] = (S_REG[PSW] & (0xFFFFFFF0|PSW_CY))|flags;
                break;
              case MOVHI:
                P_REG[arg3] = (arg1 << 16) + P_REG[arg2];
            break;
              case RETI:
                //Return from Trap/Interupt
                if(S_REG[PSW] & PSW_NP) { // Read the FE Reg
                    PC = S_REG[FEPC];
                    S_REG[PSW] = S_REG[FEPSW];
                } else { 	//Read the EI Reg Interupt
                    PC = S_REG[EIPC];
                    S_REG[PSW] = S_REG[EIPSW];
                }
                break;
              case HALT:
            dtprintf(6,ferr,"\nUnhandled opcode! halt");
            break;
          case LD_B:
            P_REG[arg3] = sign_8(mem_rbyte(sign_16(arg1)+P_REG[arg2]));
            break;
          case LD_H:
            P_REG[arg3] = sign_16(mem_rhword((sign_16(arg1)+P_REG[arg2]) & 0xFFFFFFFE));
                break;
          case MULI:
            temp= (INT64)P_REG[arg1] * (INT64)sign_16(arg3);
            //P_REG[30]   = temp >> 32;
            // Set Flags
            if ( (long)temp != temp) {
                S_REG[PSW] = S_REG[PSW] | PSW_SAT;
              }
            P_REG[arg2] = (long)temp;
            break;
          case LD_W:
            P_REG[arg3] = mem_rword((sign_16(arg1)+P_REG[arg2]) & 0xFFFFFFFC);
            break;
          case ST_B:
            mem_wbyte(sign_16(arg2)+P_REG[arg3],P_REG[arg1]&0xFF);
            break;
          case ST_H:
            mem_whword((sign_16(arg2)+P_REG[arg3])&0xFFFFFFFE,P_REG[arg1]&0xFFFF);
            break;
          case MACI:
            //dtprintf(6,ferr,"\nUnhandled opcode! maci");
            break;
          case ST_W:
            mem_wword((sign_16(arg2)+P_REG[arg3])&0xFFFFFFFC,P_REG[arg1]);
            break;
          case IN_B:
            P_REG[arg3] = port_rbyte(sign_16(arg1)+P_REG[arg2]);
            break;
          case IN_H:
            P_REG[arg3] = port_rhword((sign_16(arg1)+P_REG[arg2]) & 0xFFFFFFFE);
            break;
          case CAXI:
            //dtprintf(6,ferr,"\nUnhandled opcode! caxi");
            break;
          case IN_W:
            P_REG[arg3] = port_rword((sign_16(arg1)+P_REG[arg2]) & 0xFFFFFFFC);
            break;
          case OUT_B:
            port_wbyte(sign_16(arg2)+P_REG[arg3],P_REG[arg1]&0xFF);
            break;
          case OUT_H:
            port_whword((sign_16(arg2)+P_REG[arg3])&0xFFFFFFFE,P_REG[arg1]&0xFFFF);
            break;
          case OUT_W:
            port_wword((sign_16(arg2)+P_REG[arg3])&0xFFFFFFFC,P_REG[arg1]);
            break;
          case FPP:
            if(arg3 > 16) {
                dtprintf(6,ferr,"\n%08lx\tFPU Error: %04x", PC, arg3);
            } else {
                //dtprintf(6,ferr,"\n%08lx\t%s  $%d, $%d", PC, fpsuboptable[arg3].opname, arg1, arg2);
                (*fpsuboptable[arg3].func)(arg1, arg2, 0);    
            }
            break;
          case BSTR:
            if(arg2 > 16) {
                dtprintf(6,ferr,"\n%08lx\tBSR Error: %04x", PC,arg2);
            } else {
                //dtprintf(6,ferr,"\n%08lx\t%s, $%d", PC, bssuboptable[arg2].opname,arg1);
                (*bssuboptable[arg2].func)(arg1, 0, 0);    
            }
            break;
          default:
            break;
        }
    }
}

// Generate Interupt #n
void v810_int(WORD iNum) {
    dtprintf(4,ferr,"\nInt atempt %x",iNum);

    if (iNum > 0x0F) return;  // Invalid Interupt number...
    if((S_REG[PSW] & PSW_NP)) return;
    if((S_REG[PSW] & PSW_EP)) return; // Exception pending?
    if((S_REG[PSW] & PSW_ID)) return; // Interupt disabled
    if(iNum < ((S_REG[PSW] & PSW_IA)>>16)) return; // Interupt to low on the chain

    dtprintf(6,ferr,"\nInt %x",iNum);

    //Ready to Generate the Interupts
    S_REG[EIPC] = PC;
    S_REG[EIPSW] = S_REG[PSW];

    PC = 0xFFFFFE00 | (iNum << 4);
    
    S_REG[ECR] = 0xFE00 | (iNum << 8);
    S_REG[PSW] = S_REG[PSW] | PSW_EP;
    S_REG[PSW] = S_REG[PSW] | PSW_ID;
    if((iNum+=1) > 0x0F) (iNum = 0x0F);
    S_REG[PSW] = S_REG[PSW] | (iNum << 16); //Set the Interupt

    tVIPREG.INTPND = tVIPREG.INTENB; //Need to decode the actual Int some day... (how?)
}

/*
// Generate exception #n
//Exceptions are Div by zero, trap and Invalid Opcode, we can live without...
void v810_exp(int iNum) {
    if (iNum > 0x0F) return;  // Invalid Interupt number...
    if(!S_REG[PSW]&PSW_ID) return; 

    if(iNum < ((S_REG[PSW] & PSW_IA)>>16)) return; // Interupt to low on the mask level....
    if(S_REG[PSW]&PSW_NP) { //Fatal Exception
        S_REG[DPC] = PC;
        S_REG[DPSW] = S_REG[PSW];
        //ECR.EICC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_DP; 
        S_REG[PSW] = S_REG[PSW] | PSW_NP; 
        S_REG[PSW] = S_REG[PSW] | PSW_EP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
        
        PC = 0xFFFFFFE0;
        return;
    }else if(S_REG[PSW]&PSW_EP) { //Double Exception
        S_REG[FEPC] = PC;
        S_REG[FEPSW] = S_REG[PSW];
        //ECR.FECC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_NP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
    
        PC = 0xFFFFFFD0;
        return;
    } else {                                // Regular Exception
        S_REG[EIPC] = PC;
        S_REG[EIPSW] = S_REG[PSW];
        //ECR.EICC = Exception Code, dont get it???
        S_REG[PSW] = S_REG[PSW] | PSW_EP; 
        S_REG[PSW] = S_REG[PSW] | PSW_ID; 
        S_REG[PSW] = S_REG[PSW] | (((iNum+1) & 0x0f) << 16); //Set the Interupt status
        return;
    }
}
*/

